# SPDX-FileCopyrightText: 2019-2020 Magenta ApS
#
# SPDX-License-Identifier: MPL-2.0
"""FastAPI + RAMQP Framework."""
import logging
from contextlib import asynccontextmanager
from contextlib import AsyncExitStack
from functools import partial
from typing import Any
from typing import AsyncContextManager
from typing import AsyncGenerator
from typing import cast

import structlog
from fastapi import APIRouter
from fastapi import FastAPI
from fastapi import Request
from fastapi import Response
from prometheus_client import Info
from prometheus_fastapi_instrumentator import Instrumentator
from raclients.graph.client import GraphQLClient
from raclients.modelclient.mo import ModelClient
from ramqp.mo import MOAMQPSystem
from starlette.status import HTTP_204_NO_CONTENT
from starlette.status import HTTP_503_SERVICE_UNAVAILABLE

from .config import Settings
from .context import Context
from .context import HealthcheckFunction
from .healthcheck import healthcheck_gql
from .healthcheck import healthcheck_model_client


logger = structlog.get_logger()
fastapi_router = APIRouter()
build_information = Info("build_information", "Build information")


def update_build_information(version: str, build_hash: str) -> None:
    """Update build information.

    Args:
        version: The version to set.
        build_hash: The build hash to set.

    Returns:
        None.
    """
    build_information.info(
        {
            "version": version,
            "hash": build_hash,
        }
    )


def configure_logging(log_level_name: str) -> None:
    """Setup our logging.

    Args:
        log_level_name: The logging level.

    Returns:
        None
    """
    log_level_value = logging.getLevelName(log_level_name)
    structlog.configure(
        wrapper_class=structlog.make_filtering_bound_logger(log_level_value)
    )


@fastapi_router.get("/")
async def index(request: Request) -> dict[str, str]:
    """Endpoint to return name of integration."""
    context: dict[str, Any] = request.app.state.context
    return {"name": context["name"]}


@fastapi_router.get("/health/live", status_code=HTTP_204_NO_CONTENT)
async def liveness() -> None:
    """Endpoint to be used as a liveness probe for Kubernetes."""
    return None


@fastapi_router.get(
    "/health/ready",
    status_code=HTTP_204_NO_CONTENT,
    responses={
        "204": {"description": "Ready"},
        "503": {"description": "Not ready"},
    },
)
async def readiness(request: Request, response: Response) -> Response:
    """Endpoint to be used as a readiness probe for Kubernetes."""
    response.status_code = HTTP_204_NO_CONTENT

    context: dict[str, Any] = request.app.state.context
    healthchecks = context["healthchecks"]
    all_ready = True
    try:
        for name, healthcheck in healthchecks.items():
            ready = await healthcheck(context)
            if not ready:
                logger.warn(f"{name} is not ready")
                all_ready = False
    except Exception:  # pylint: disable=broad-except
        logger.exception("Exception occured during readiness probe")
        response.status_code = HTTP_503_SERVICE_UNAVAILABLE

    if not all_ready:
        response.status_code = HTTP_503_SERVICE_UNAVAILABLE

    return response


def construct_clients(
    settings: Settings,
) -> tuple[GraphQLClient, ModelClient]:
    """Construct clients froms settings.

    Args:
        settings: Integration settings module.

    Returns:
        Tuple with PersistentGraphQLClient and ModelClient.
    """
    client_kwargs = dict(
        client_id=settings.client_id,
        client_secret=settings.client_secret.get_secret_value(),
        auth_realm=settings.auth_realm,
        auth_server=settings.auth_server,
    )

    gql_client = GraphQLClient(
        url=settings.mo_url + "/graphql",
        execute_timeout=settings.graphql_timeout,
        httpx_client_kwargs={"timeout": settings.graphql_timeout},
        **client_kwargs,
    )
    model_client = ModelClient(
        base_url=settings.mo_url,
        **client_kwargs,
    )
    return gql_client, model_client


@asynccontextmanager
async def _lifespan(_1: FastAPI, context: Context) -> AsyncGenerator[None, None]:
    """ASGI lifespan context handler.

    Runs all the configured lifespan managers according to their priority.

    Returns:
        None
    """
    async with AsyncExitStack() as stack:
        lifespan_managers = context["lifespan_managers"]
        for _, priority_set in sorted(lifespan_managers.items()):
            for lifespan_manager in priority_set:
                await stack.enter_async_context(lifespan_manager)
        # Yield to keep lifespan managers open until the ASGI application is shutdown.
        yield


class FastRAMQPI:
    """FastRAMQPI (FastAPI + RAMQP) combined-system.

    Motivated by a lot a shared code between our AMQP integrations.
    """

    def __init__(self, application_name: str, settings: Settings | None = None) -> None:
        super().__init__()
        if settings is None:
            settings = Settings()
        configure_logging(settings.log_level)

        # Update metrics info
        update_build_information(
            version=settings.commit_tag, build_hash=settings.commit_sha
        )

        # Setup shared context
        self._context: Context = {
            "name": application_name,
            "settings": settings,
            "healthchecks": {},
            "lifespan_managers": {},
            "user_context": {},
        }

        # Setup FastAPI
        app = FastAPI()
        app.include_router(fastapi_router)
        app.state.context = self._context
        app.router.lifespan_context = partial(_lifespan, context=self._context)
        # Expose Metrics
        if settings.enable_metrics:
            Instrumentator().instrument(app).expose(app)
        self.app = app
        self._context["app"] = self.app

        # Setup AMQPSystem
        amqp_settings = settings.amqp.copy(update={"queue_prefix": application_name})
        self.amqpsystem = MOAMQPSystem(settings=amqp_settings, context=self._context)
        # Let AMQPSystems lifespan follow ASGI lifespan
        self.add_lifespan_manager(self.amqpsystem)

        async def healthcheck_amqp(context: Context) -> bool:
            """AMQP Healthcheck wrapper.

            Args:
                context: unused context dict.

            Returns:
                Whether the AMQPSystem is OK.
            """
            amqpsystem = context["amqpsystem"]
            return cast(bool, amqpsystem.healthcheck())

        self.add_healthcheck(name="AMQP", healthcheck=healthcheck_amqp)
        self._context["amqpsystem"] = self.amqpsystem

        # Prepare clients
        graphql_client, model_client = construct_clients(settings)
        # Check and expose GraphQL connection (gql_client)
        self.add_healthcheck("GraphQL", healthcheck_gql)
        self._context["graphql_client"] = graphql_client

        @asynccontextmanager
        async def graphql_session(context: Context) -> AsyncGenerator[None, None]:
            async with context["graphql_client"] as session:
                context["graphql_session"] = session
                yield

        self.add_lifespan_manager(
            cast(AsyncContextManager, partial(graphql_session, self._context)())
        )
        # Check and expose Service API connection (model_client)
        self.add_lifespan_manager(model_client)
        self.add_healthcheck("Service API", healthcheck_model_client)
        self._context["model_client"] = model_client

    def add_lifespan_manager(
        self, manager: AsyncContextManager, priority: int = 1000
    ) -> None:
        """Add the provided life-cycle manager to the ASGI lifespan context.

        Args:
            manager: The manager to add.
            priority: The priority of the manager, lowest priorities are run first.

        Returns:
            None
        """

        priority_set = self._context["lifespan_managers"].setdefault(priority, set())
        priority_set.add(manager)

    def add_healthcheck(self, name: str, healthcheck: HealthcheckFunction) -> None:
        """Add the provided healthcheck to the Kubernetes readiness probe.

        Args:
            name: Name of the healthcheck to add.
            healthcheck: The healthcheck callback function.

        Raises:
            ValueError: If the name has already been used.

        Returns:
            None
        """
        if name in self._context["healthchecks"]:
            raise ValueError("Name already used")
        self._context["healthchecks"][name] = healthcheck

    def add_context(self, **kwargs: Any) -> None:
        """Add the provided key-value pair to the user-context.

        The added key-value pair will be available under context["user_context"].

        Args:
            key: The key to add under.
            value: The value to add.

        Returns:
            None
        """
        self._context["user_context"].update(**kwargs)

    def get_context(self) -> Context:
        """Return the contained context.

        Returns:
            The contained context.
        """
        return self._context

    def get_app(self) -> FastAPI:
        """Return the contained FastAPI application.

        Returns:
            FastAPI application.
        """
        return self.app

    def get_amqpsystem(self) -> MOAMQPSystem:
        """Return the contained MOAMQPSystem.

        Returns:
            MOAQMPSystem.
        """
        return self.amqpsystem



# SPDX-FileCopyrightText: 2019-2020 Magenta ApS
#
# SPDX-License-Identifier: MPL-2.0
"""Healthchecking functions."""
import structlog
from gql import gql
from more_itertools import one

from .context import Context


logger = structlog.get_logger()


async def healthcheck_gql(context: Context) -> bool:
    """Check that our GraphQL connection is healthy.

    Args:
        context: Execution context that contains GraphQL client.

    Returns:
        Whether the client is healthy or not.
    """
    gql_client = context["graphql_session"]

    query = gql(
        """
        query HealthcheckQuery {
            org {
                uuid
            }
        }
        """
    )
    try:
        result = await gql_client.execute(query)
        if result["org"]["uuid"]:
            return True
    except Exception:  # pylint: disable=broad-except
        logger.exception("Exception occured during GraphQL healthcheck")
    return False


async def healthcheck_model_client(context: Context) -> bool:
    """Check that our ModelClient connection is healthy.

    Args:
        context: Execution context that contains MO model client.

    Returns:
        Whether the client is healthy or not.
    """
    model_client = context["model_client"]

    try:
        response = await model_client.get("/service/o/")
        result = response.json()
        if one(result)["uuid"]:
            return True
    except Exception:  # pylint: disable=broad-except
        logger.exception("Exception occured during GraphQL healthcheck")
    return False



# SPDX-FileCopyrightText: 2021 Magenta ApS <https://magenta.dk>
# SPDX-License-Identifier: MPL-2.0
"""Settings handling."""
from pydantic import AnyHttpUrl
from pydantic import BaseSettings
from pydantic import Field
from pydantic import parse_obj_as
from pydantic import SecretStr
from ramqp.config import ConnectionSettings


# pylint: disable=too-few-public-methods
class Settings(BaseSettings):
    """Settings for the FastRAMQPI framework."""

    class Config:
        """Settings are frozen."""

        frozen = True
        env_nested_delimiter = "__"

    amqp: ConnectionSettings = Field(
        ConnectionSettings(), description="AMQP connection settings"
    )

    # We assume these will be set by the docker build process,
    # and as such will contain release information at runtime.
    commit_tag: str = Field("HEAD", description="Git commit tag.")
    commit_sha: str = Field("HEAD", description="Git commit SHA.")

    log_level: str = Field("INFO", description="Log level to configure.")

    enable_metrics: bool = Field(True, description="Whether to enable metrics.")

    mo_url: AnyHttpUrl = Field(
        parse_obj_as(AnyHttpUrl, "http://mo-service:5000"),
        description="Base URL for OS2mo.",
    )
    client_id: str = Field("orggatekeeper", description="Client ID for OIDC client.")
    client_secret: SecretStr = Field(..., description="Client Secret for OIDC client.")
    auth_server: AnyHttpUrl = Field(
        parse_obj_as(AnyHttpUrl, "http://keycloak-service:8080/auth"),
        description="Base URL for OIDC server (Keycloak).",
    )
    auth_realm: str = Field("mo", description="Realm to authenticate against")
    graphql_timeout: int = 120



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['fastramqpi']

package_data = \
{'': ['*']}

install_requires = \
['fastapi>=0.78.0,<0.79.0',
 'gql>=3.3.0,<4.0.0',
 'more-itertools>=8.13.0,<9.0.0',
 'prometheus-fastapi-instrumentator>=5.8.1,<6.0.0',
 'pydantic>=1.9.0,<2.0.0',
 'ra-utils>=1.0.0,<2.0.0',
 'raclients>=1.2.5,<2.0.0',
 'ramqp>=6.0.0,<7.0.0',
 'rautils>=1.0.10,<2.0.0',
 'structlog>=21.5.0,<22.0.0']

setup_kwargs = {
    'name': 'fastramqpi',
    'version': '1.0.0',
    'description': 'Rammearkitektur AMQP framework (FastAPI + RAMQP)',
    'long_description': '<!--\nSPDX-FileCopyrightText: 2021 Magenta ApS <https://magenta.dk>\nSPDX-License-Identifier: MPL-2.0\n-->\n\n# FastRAMQPI\n\nFastRAMQPI is an opinionated library for FastAPI and RAMQP.\n\nIt is implemented as a thin wrapper around `FastAPI` and `RAMQP`.\nIt is very MO specific.\n\n## Usage\n\n```python\nfrom pydantic import BaseSettings\nfrom fastramqpi import FastRAMQPI\nfrom fastramqpi import FastRAMQPISettings\n\n\nclass Settings(BaseSettings):\n    class Config:\n        frozen = True\n        env_nested_delimiter = "__"\n\n    fastramqpi: FastRAMQPISettings = Field(\n        default_factory=FastRAMQPISettings,\n        description="FastRAMQPI settings"\n    )\n\n    # All your program settings hereunder...\n\n\nfastapi_router = APIRouter()\n\n@fastapi_router.post("/trigger/all")\nasync def update_all(request: Request) -> dict[str, str]:\n    context: dict[str, Any] = request.app.state.context\n    graphql_session = context["grapqh_session"]\n    program_settings = context["user_context"]["settings"]\n    ...\n    return {"status": "OK"}\n\n\namqp_router = MORouter()\n\n@amqp_router.register("*.*.*")\nasync def listen_to_all(context: dict, payload: PayloadType) -> None:\n    graphql_session = context["grapqh_session"]\n    program_settings = context["user_context"]["settings"]\n    ...\n\n\ndef create_fastramqpi(**kwargs: Any) -> FastRAMQPI:\n    settings = Settings(**kwargs)\n    fastramqpi = FastRAMQPI(\n        application_name="orggatekeeper", settings=settings.fastramqpi\n    )\n    fastramqpi.add_context(settings=settings)\n\n    # Add our AMQP router(s)\n    amqpsystem = fastramqpi.get_amqpsystem()\n    amqpsystem.router.registry.update(amqp_router.registry)\n\n    # Add our FastAPI router(s)\n    app = fastramqpi.get_app()\n    app.include_router(fastapi_router)\n\n    return fastramqpi\n\n\ndef create_app(**kwargs: Any) -> FastAPI:\n    fastramqpi = create_fastramqpi(**kwargs)\n    return fastramqpi.get_app()\n```\n\n### Metrics\nFastRAMQPI Metrics are exported via `prometheus/client_python` on the FastAPI\'s `/metrics`.\n\n## Development\n\n### Prerequisites\n\n- [Poetry](https://github.com/python-poetry/poetry)\n\n### Getting Started\n\n1. Clone the repository:\n```\ngit clone git@git.magenta.dk:rammearkitektur/FastRAMQPI.git\n```\n\n2. Install all dependencies:\n```\npoetry install\n```\n\n3. Set up pre-commit:\n```\npoetry run pre-commit install\n```\n\n### Running the tests\n\nYou use `poetry` and `pytest` to run the tests:\n\n`poetry run pytest`\n\nYou can also run specific files\n\n`poetry run pytest tests/<test_folder>/<test_file.py>`\n\nand even use filtering with `-k`\n\n`poetry run pytest -k "Manager"`\n\nYou can use the flags `-vx` where `v` prints the test & `x` makes the test stop if any tests fails (Verbose, X-fail)\n\n## Authors\n\nMagenta ApS <https://magenta.dk>\n\n## License\n\nThis project uses: [MPL-2.0](LICENSES/MPL-2.0.txt)\n\nThis project uses [REUSE](https://reuse.software) for licensing.\nAll licenses can be found in the [LICENSES folder](LICENSES/) of the project.\n',
    'author': 'Magenta ApS',
    'author_email': 'info@magenta.dk',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://magenta.dk/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
